# Copyright 2001 by Tarjei Mikkelsen. All rights reserved.
"""BioPython Pathway support module.

Bio.Pathway.Rep is a collection of general data abstractions used in the
implementation of the Bio.Pathway module. These abstractions are not intended
for direct use, but if they fit your needs there's no reason to reinvent the
wheel.
"""
