# Copyright 2001 by Brad Chapman.  All rights reserved.
#
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.

"""Back-propagation for neural networks."""
# TODO: Remove empty __init__.py once we drop Python 2 support
